
/*

    Nn2GS - Übersetzt zwischen Normalnoten und Griffschrift-Tabulatur für Steirische Harmonika.
    Copyright (C) 2021  Jakob Schöttl <jschoett@gmail.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

*/

import QtQuick 2.9
import MuseScore 3.0
import QtQuick.Dialogs 1.2
import QtQuick.Controls 2.2
import QtQuick.Layouts 1.2

// Discussion about QtQuick versions:
// https://musescore.org/en/node/281629
// https://doc.qt.io/qt-5/qtquickcontrols-index.html#versions

// Funktionen, die ich vllt noch brauche:
// newElement(type: int)
// removeElement(elem)
// openLog()
// closeLog()
// log(), logn(), log2()

MuseScore {
    id: plugin
    menuPath:   "Plugins.Griffschrift (Nn2GS)"
version: "1.1.2"
    description: "Dieses Plugin übersetzt zwischen Normalnoten und Griffschrift-Tabulatur für Steirische Harmonika."
    requiresScore: true

    pluginType: "dock"
    dockArea:   "left"

    implicitWidth:  200
    implicitHeight: 700
    //width:  200
    //height: 200
    //Layout.fillWidth: true

    // Polyfills used by compiled functions:
    property var $jscomp: {}

    onRun: {
        console.log("Starting...")
        btnReverseDirection.text = btnReverseDirection.texts[btnReverseDirection.state]
        btnReverseDirection.background.color = colorButtonNormal

        $jscomp = {} // Polyfill object
        // POLYFILL IMPLEMENTATION HERE:
$jscomp.scope = {};
$jscomp.createTemplateTagFirstArg = function(a) {
  return a.raw = a;
};
$jscomp.createTemplateTagFirstArgWithRaw = function(a, b) {
  a.raw = b;
  return a;
};
$jscomp.arrayIteratorImpl = function(a) {
  var b = 0;
  return function() {
    return b < a.length ? {done:!1, value:a[b++], } : {done:!0};
  };
};
$jscomp.arrayIterator = function(a) {
  return {next:$jscomp.arrayIteratorImpl(a)};
};
$jscomp.makeIterator = function(a) {
  var b = "undefined" != typeof Symbol && Symbol.iterator && a[Symbol.iterator];
  return b ? b.call(a) : $jscomp.arrayIterator(a);
};
    }
    onScoreStateChanged: {
        if (state.selectionChanged) {
            // Invalidate current translation info and selection
            console.log("Selection changed")
            invalidateCurrentResults()
            // Selection changes after converting GS → Nn; therefore
            // the disableZDButtonsForTimeout wouldn't have an effect.
            //enableZDButtons(true)
        }
    }

    //readonly property string apiUrl: "https://ziach.intensovet.de/nn2gs"
    readonly property string apiUrl: "https://ziach.intensovet.de/nn2gs"

    readonly property var colorRed: "#ff0000"
    readonly property var colorButtonNormal: "#bbb"
    readonly property var colorDirectionReverse: "#666"

    // Generated using Nn2GS.hs:
    // result = map (\(x,y) -> (x, map toLower.noteNameToGerman.show $ y, computeAccidentals $ Dur y)) quintenZirkel
    // encode result
    // putStrLn ...
    readonly property var tonarten: [[-8,"fes",["Bes","Es","As","Des","Ges","Ces","Fes"]],[-7,"ces",["Bes","Es","As","Des","Ges","Ces","Fes"]],[-6,"ges",["Bes","Es","As","Des","Ges","Ces"]],[-5,"des",["Bes","Es","As","Des","Ges"]],[-4,"as",["Bes","Es","As","Des"]],[-3,"es",["Bes","Es","As"]],[-2,"b",["Bes","Es"]],[-1,"f",["Bes"]],[0,"c",[]],[1,"g",["Fis"]],[2,"d",["Fis","Cis"]],[3,"a",["Fis","Cis","Gis"]],[4,"e",["Fis","Cis","Gis","Dis"]],[5,"h",["Fis","Cis","Gis","Dis","Ais"]],[6,"fis",["Fis","Cis","Gis","Dis","Ais","Eis"]],[7,"cis",["Fis","Cis","Gis","Dis","Ais","Eis","Bis"]],[8,"gis",["Fis","Cis","Gis","Dis","Ais","Eis","Bis"]],[9,"dis",["Fis","Cis","Gis","Dis","Ais","Eis","Bis"]],[10,"ais",["Fis","Cis","Gis","Dis","Ais","Eis","Bis"]],[11,"eis",["Fis","Cis","Gis","Dis","Ais","Eis","Bis"]],[12,"his",["Fis","Cis","Gis","Dis","Ais","Eis","Bis"]]]

    // Generated using Nn2GS.hs:
    // Exclude some notes; see documentation in getNoteName().
    // encode . filter (not . flip elem [Eis, Bis, Ces, Fes] . withoutOctave . fst) . M.toList $ midiPitchMap
    // putStrLn "..."
    readonly property var midiPitchMap: [["C_",36],["Cis_",37],["Des_",37],["D_",38],["Dis_",39],["Es_",39],["E_",40],["F_",41],["Fis_",42],["Ges_",42],["G_",43],["Gis_",44],["As_",44],["A_",45],["Ais_",46],["Bes_",46],["B_",47],["C",48],["Cis",49],["Des",49],["D",50],["Dis",51],["Es",51],["E",52],["F",53],["Fis",54],["Ges",54],["G",55],["Gis",56],["As",56],["A",57],["Ais",58],["Bes",58],["B",59],["C'",60],["Cis'",61],["Des'",61],["D'",62],["Dis'",63],["Es'",63],["E'",64],["F'",65],["Fis'",66],["Ges'",66],["G'",67],["Gis'",68],["As'",68],["A'",69],["Ais'",70],["Bes'",70],["B'",71],["C''",72],["Cis''",73],["Des''",73],["D''",74],["Dis''",75],["Es''",75],["E''",76],["F''",77],["Fis''",78],["Ges''",78],["G''",79],["Gis''",80],["As''",80],["A''",81],["Ais''",82],["Bes''",82],["B''",83],["C'''",84],["Cis'''",85],["Des'''",85],["D'''",86],["Dis'''",87],["Es'''",87],["E'''",88],["F'''",89],["Fis'''",90],["Ges'''",90],["G'''",91],["Gis'''",92],["As'''",92],["A'''",93],["Ais'''",94],["Bes'''",94],["B'''",95],["C''''",96],["Cis''''",97],["Des''''",97],["D''''",98],["Dis''''",99],["Es''''",99],["E''''",100],["F''''",101],["Fis''''",102],["Ges''''",102],["G''''",103],["Gis''''",104],["As''''",104],["A''''",105],["Ais''''",106],["Bes''''",106],["B''''",107]]

    // Generated using Nn2GS.hs:
    // map (\x -> [show x, map toLower . noteNameToGerman . show $ x]) $ enumFrom C_
    readonly property var germanNoteNames: [["C_","c_"],["Cis_","cis_"],["Des_","des_"],["D_","d_"],["Dis_","dis_"],["Es_","es_"],["E_","e_"],["Fes_","fes_"],["Eis_","eis_"],["F_","f_"],["Fis_","fis_"],["Ges_","ges_"],["G_","g_"],["Gis_","gis_"],["As_","as_"],["A_","a_"],["Ais_","ais_"],["Bes_","b_"],["B_","h_"],["Ces","ces"],["Bis_","his_"],["C","c"],["Cis","cis"],["Des","des"],["D","d"],["Dis","dis"],["Es","es"],["E","e"],["Fes","fes"],["Eis","eis"],["F","f"],["Fis","fis"],["Ges","ges"],["G","g"],["Gis","gis"],["As","as"],["A","a"],["Ais","ais"],["Bes","b"],["B","h"],["Ces'","ces'"],["Bis","his"],["C'","c'"],["Cis'","cis'"],["Des'","des'"],["D'","d'"],["Dis'","dis'"],["Es'","es'"],["E'","e'"],["Fes'","fes'"],["Eis'","eis'"],["F'","f'"],["Fis'","fis'"],["Ges'","ges'"],["G'","g'"],["Gis'","gis'"],["As'","as'"],["A'","a'"],["Ais'","ais'"],["Bes'","b'"],["B'","h'"],["Ces''","ces''"],["Bis'","his'"],["C''","c''"],["Cis''","cis''"],["Des''","des''"],["D''","d''"],["Dis''","dis''"],["Es''","es''"],["E''","e''"],["Fes''","fes''"],["Eis''","eis''"],["F''","f''"],["Fis''","fis''"],["Ges''","ges''"],["G''","g''"],["Gis''","gis''"],["As''","as''"],["A''","a''"],["Ais''","ais''"],["Bes''","b''"],["B''","h''"],["Ces'''","ces'''"],["Bis''","his''"],["C'''","c'''"],["Cis'''","cis'''"],["Des'''","des'''"],["D'''","d'''"],["Dis'''","dis'''"],["Es'''","es'''"],["E'''","e'''"],["Fes'''","fes'''"],["Eis'''","eis'''"],["F'''","f'''"],["Fis'''","fis'''"],["Ges'''","ges'''"],["G'''","g'''"],["Gis'''","gis'''"],["As'''","as'''"],["A'''","a'''"],["Ais'''","ais'''"],["Bes'''","b'''"],["B'''","h'''"],["Ces''''","ces''''"],["Bis'''","his'''"],["C''''","c''''"],["Cis''''","cis''''"],["Des''''","des''''"],["D''''","d''''"],["Dis''''","dis''''"],["Es''''","es''''"],["E''''","e''''"],["Fes''''","fes''''"],["Eis''''","eis''''"],["F''''","f''''"],["Fis''''","fis''''"],["Ges''''","ges''''"],["G''''","g''''"],["Gis''''","gis''''"],["As''''","as''''"],["A''''","a''''"],["Ais''''","ais''''"],["Bes''''","b''''"],["B''''","h''''"],["Ces'''''","ces'''''"],["Bis''''","his''''"]]

    // Generated using Nn2GS.hs:
    // tonalPitches = zip [-1::Int ..4] $ repeat $ flip zip [6::Int ..] $ map snd quintenZirkel
    // encode $ concatMap (\(i, xs) -> map (\(x, p) -> (show $ mkTon' x i, p)) xs) $ map (\(i, xs) -> (i, filter (\(x,_) -> not (i == -1 && x == Ces)) xs)) $ tonalPitches
    // putStrLn ...
    readonly property var tonalPitchClassMap: [["Fes_",6],["Ges_",8],["Des_",9],["As_",10],["Es_",11],["Bes_",12],["F_",13],["C_",14],["G_",15],["D_",16],["A_",17],["E_",18],["B_",19],["Fis_",20],["Cis_",21],["Gis_",22],["Dis_",23],["Ais_",24],["Eis_",25],["Bis_",26],["Fes",6],["Ces",7],["Ges",8],["Des",9],["As",10],["Es",11],["Bes",12],["F",13],["C",14],["G",15],["D",16],["A",17],["E",18],["B",19],["Fis",20],["Cis",21],["Gis",22],["Dis",23],["Ais",24],["Eis",25],["Bis",26],["Fes'",6],["Ces'",7],["Ges'",8],["Des'",9],["As'",10],["Es'",11],["Bes'",12],["F'",13],["C'",14],["G'",15],["D'",16],["A'",17],["E'",18],["B'",19],["Fis'",20],["Cis'",21],["Gis'",22],["Dis'",23],["Ais'",24],["Eis'",25],["Bis'",26],["Fes''",6],["Ces''",7],["Ges''",8],["Des''",9],["As''",10],["Es''",11],["Bes''",12],["F''",13],["C''",14],["G''",15],["D''",16],["A''",17],["E''",18],["B''",19],["Fis''",20],["Cis''",21],["Gis''",22],["Dis''",23],["Ais''",24],["Eis''",25],["Bis''",26],["Fes'''",6],["Ces'''",7],["Ges'''",8],["Des'''",9],["As'''",10],["Es'''",11],["Bes'''",12],["F'''",13],["C'''",14],["G'''",15],["D'''",16],["A'''",17],["E'''",18],["B'''",19],["Fis'''",20],["Cis'''",21],["Gis'''",22],["Dis'''",23],["Ais'''",24],["Eis'''",25],["Bis'''",26],["Fes''''",6],["Ces''''",7],["Ges''''",8],["Des''''",9],["As''''",10],["Es''''",11],["Bes''''",12],["F''''",13],["C''''",14],["G''''",15],["D''''",16],["A''''",17],["E''''",18],["B''''",19],["Fis''''",20],["Cis''''",21],["Gis''''",22],["Dis''''",23],["Ais''''",24],["Eis''''",25],["Bis''''",26]]


    // Prevent memory/performance problems and API abuse
    readonly property int maxChordLimit: 50

    property var lastResults: null
    property string lastZD: ''
    property int alternativeIndex: 0

    // PURE FUNCTIONS HERE:
function jsonStringifyNonRecursive(a) {
  var b = [];
  return JSON.stringify(a, function(c, d) {
    if ("object" === typeof d && null !== d) {
      if (b.includes(d)) {
        return;
      }
      b.push(d);
    }
    return d;
  });
}
function propertiesOfObject(a) {
  var b = [], c;
  for (c in a) {
    b.push(c);
  }
  return b;
}
function setAccidentalVisible(a, b) {
  console.log(a.accidental);
  a.accidental && (console.log(jsonStringifyNonRecursive(a.accidental)), a.accidental.visible = b);
}
function addSharp2(a) {
  var b = a.pitch;
  a.tpc += 14;
  a.pitch = b + 2;
  a.accidentalType = Accidental.SHARP2;
  a.accidental && (a.accidental.offsetX = 0);
}
function removeSharp2(a) {
  var b = a.pitch;
  a.tpc -= 14;
  a.pitch = b - 2;
  a.accidentalType = Accidental.NONE;
}
function durationIs(a, b) {
  return 1 === a.duration.numerator && a.duration.denominator === b;
}
function hasSpecialNoteHead(a) {
  return a.accidentalType == Accidental.SHARP2 || a.headGroup != NoteHeadGroup.HEAD_NORMAL;
}
function hasCrossedNoteHead(a) {
  return a.accidentalType == Accidental.SHARP2 || a.headGroup == NoteHeadGroup.HEAD_CROSS || a.headGroup == NoteHeadGroup.HEAD_XCIRCLE;
}
function looksLikeGriffschrift(a) {
  return a.map(function(b) {
    for (var c = 0; c < b.notes.length; c++) {
      if (hasSpecialNoteHead(b.notes[c])) {
        return !0;
      }
    }
    return !1;
  }).some(function(b) {
    return b;
  });
}
function queryStringArg(a, b, c) {
  return ((void 0 === c ? 0 : c) ? "" : "&") + (encodeURIComponent(a) + "=" + encodeURIComponent(b));
}
function chordsAsApiInput(a, b) {
  var c = b ? function(d, e) {
    d.accidentalType == Accidental.SHARP2 && (e -= 2);
    return {crossed:hasCrossedNoteHead(d), position:getNoteName(e)};
  } : function(d, e) {
    return getNoteName(e);
  };
  return a.map(function(d) {
    var e = [];
    d = d.notes;
    for (var f = 0; f < d.length; f++) {
      var g = d[f];
      e.push(c(g, g.pitch));
    }
    console.log(e);
    return e;
  });
}
function getMidiPitch(a) {
  var b = midiPitchMap.find(function(c) {
    c = $jscomp.makeIterator(c);
    var d = c.next().value;
    c.next();
    return d === a;
  });
  return b ? b[1] : null;
}
function getNoteName(a) {
  var b = midiPitchMap.find(function(c) {
    c = $jscomp.makeIterator(c);
    c.next();
    return c.next().value === a;
  });
  return b ? b[0] : null;
}
function extractZDResults(a, b) {
  return b["druck" === a ? 1 : 0];
}
function lookupTonalPitchClass(a) {
  return tonalPitchClassMap.find(function(b) {
    b = $jscomp.makeIterator(b);
    var c = b.next().value;
    b.next();
    return a == c;
  })[1];
}
function containsRedNote(a) {
  for (var b = 0; b < a.length; b++) {
    if (a[b].color == colorRed) {
      return !0;
    }
  }
  return !1;
}
function constructChordsFromNormalResults(a) {
  return a.map(function(b) {
    var c = [];
    b.Right ? c = b.Right.map(function(d) {
      return {pitch:getMidiPitch(d)};
    }) : b.Left ? (console.warn("Did not receive expected normal notation for current chord: " + JSON.stringify(result)), c = b.Left.map(function(d) {
      return {pitch:getMidiPitch(d.position)};
    })) : console.warn("Invalid result for current chord: " + JSON.stringify(result));
    return {notes:c};
  });
}


    // IMPURE FUNCTIONS HERE:
function applyToChordsInSelection(c, a) {
  var b = curScore.newCursor();
  b.rewind(1);
  var e = !1;
  if (b.segment) {
    var d = b.staffIdx;
    b.rewind(2);
    var f = 0 === b.tick ? curScore.lastSegment.tick + 1 : b.tick;
    var h = b.staffIdx;
  } else {
    e = !0, d = 0, h = curScore.nstaves - 1;
  }
  console.log("applyToChordsInSelection: " + d + " - " + h + " - " + f);
  for (var g = 0; d <= h; d++) {
    for (var k = 0; 4 > k; k++) {
      for (b.rewind(1), b.voice = k, b.staffIdx = d, e && b.rewind(0); b.segment && (e || b.tick < f);) {
        if (b.element && b.element.type == Element.CHORD) {
          for (var m = b.element.graceNotes, l = 0; l < m.length; l++) {
            if (a(m[l]), g++, g >= c) {
              return;
            }
          }
          a(b.element);
          g++;
          if (g >= c) {
            return;
          }
        }
        b.next();
      }
    }
  }
}
function callApi(c, a, b) {
  var e = JSON.stringify(chordsAsApiInput(c, a));
  a = "?" + queryStringArg("tonart", tonarten[spinnerTonart.value][1], !0) + queryStringArg("model", comboModel.currentKey()) + (a ? queryStringArg("reverse", "yes") : "") + (txtLicenseKey.text ? queryStringArg("license", txtLicenseKey.text) : "");
  console.log(a);
  var d = new XMLHttpRequest;
  d.onreadystatechange = function() {
    if (d.readyState === XMLHttpRequest.DONE) {
      if (console.log("HTTP status: " + d.status), 200 === d.status) {
        console.log("API response:\n" + d.responseText);
        try {
          lastResults = JSON.parse(d.responseText);
        } catch (f) {
          console.error("Invalid JSON response. Could not parse."), errorDialog.show("Ung\u00fcltige Antwort vom Server.");
        }
        b(c, lastResults);
      } else {
        500 <= d.status ? errorDialog.show("Fehler beim Server. Funktioniert " + apiUrl + "?") : 400 <= d.status ? errorDialog.show("Fehler bei der Kommunikation mit dem Server. Funktioniert " + apiUrl + "?") : 300 <= d.status && errorDialog.show("Wahrscheinlich passt die Lizenz nicht. \u00dcberpr\u00fcfen auf " + apiUrl + ".");
      }
    } else {
      console.log("HTTP request ready status: " + d.readyState + " (not DONE)");
    }
  };
  console.log("Run in shell for testing:\ncat <<TEXT > test.json\n" + e + '\nTEXT\ncurl -H "Content-Type: application/json" --data-binary @test.json "' + (apiUrl + a) + '"\n');
  d.open("POST", apiUrl + a, !0);
  d.setRequestHeader("Content-Type", "application/json");
  d.send(e);
}
function collectChords() {
  var c = [];
  applyToChordsInSelection(maxChordLimit + 1, function(a) {
    containsRedNote(a) ? console.log("Collecting chords: Current chord contains a red note. Skipping to next chord.") : c.push(a);
  });
  return c;
}
function changeNotes(c, a) {
  return function(b, e) {
    curScore.startCmd();
    lastZD = c;
    e = extractZDResults(c, e);
    console.log(c + " " + e);
    if (b.length !== e.length) {
      console.warn("Length of selected chords (" + b.length + ") and translation result (" + e.length + ") do not match. Aborting.");
    } else {
      for (var d = a ? changeNotesOfChordReverse : changeNotesOfChord, f = 0; f < b.length; f++) {
        d(b[f], e[f], f);
      }
      curScore.endCmd();
    }
  };
}
function changeNotesOfChordReverse(c, a, b) {
  c = c.notes;
  if (containsRedNote(c)) {
    console.log("Current chord " + b + " contains a red note. Leaving it unchanged. Skipping to next chord.");
  } else {
    if (a.Right) {
      a = a.Right;
      for (var e = 0; e < c.length; e++) {
        var d = a[e], f = c[e];
        if (c.length !== a.length) {
          console.warn("Length of current chord (" + c.length + ") and translation result (" + a.length + ") do not match in chord " + b + ", note " + e + ". Skipping to next chord.");
          break;
        }
        var h = getMidiPitch(d);
        null === h ? console.warn("Invalid note position " + d + " in translation result in chord " + b + ", note " + e + ". Skipping to next note.") : (d = lookupTonalPitchClass(d), f.accidentalType = Accidental.NONE, f.headGroup = NoteHeadGroup.HEAD_NORMAL, f.mirrorHead = 0, console.log("Changing " + f.pitch + " (tpc=" + f.tpc + ") to " + h + " (tpc=" + d + ")"), f.pitch = h, f.tpc = d, setAccidentalVisible(!0));
      }
    } else {
      if (a.Left) {
        for (b = a.Left.map(function(g) {
          var k = g.crossed;
          return [getMidiPitch(g.position), k];
        }), a = 0; a < c.length; a++) {
          e = c[a], b.includes([e.pitch, e.headGroup == NoteHeadGroup.HEAD_CROSS]) && (e.color = colorRed);
        }
      } else {
        console.warn("Invalid result for current chord " + b + ": " + JSON.stringify(a));
      }
    }
  }
}
function changeNotesOfChord(c, a, b) {
  var e = c.notes;
  if (containsRedNote(e)) {
    console.log("Current chord " + b + " contains a red note. Leaving it unchanged. Skipping to next chord.");
  } else {
    if (a.Right) {
      for (var d = 0; d < e.length; d++) {
        var f = a.Right[alternativeIndex % a.Right.length], h = f[d], g = e[d];
        if (e.length !== f.length) {
          console.warn("Length of current chord (" + e.length + ") and translation result (" + f.length + ") do not match in chord " + b + ", note " + d + ". Skipping to next chord.");
          break;
        }
        f = getMidiPitch(h.position);
        if (null === f) {
          console.warn("Invalid note position " + h.position + " in translation result in chord " + b + ", note " + d + ". Skipping to next note.");
        } else {
          var k = lookupTonalPitchClass(h.position);
          console.log("Changing " + g.pitch + " (tpc=" + g.tpc + ") to " + f + " (tpc=" + k + ")");
          g.pitch = f;
          g.tpc = k;
          null === h.side ? g.mirrorHead = 0 : "Links" === h.side ? g.mirrorHead = 1 : "Rechts" === h.side && (g.mirrorHead = 2);
          g.headGroup = NoteHeadGroup.HEAD_NORMAL;
          setAccidentalVisible(!1);
          h.extra && 0 === h.row ? g.headGroup = NoteHeadGroup.HEAD_TRIANGLE_UP : h.extra && 1 === h.row ? g.headGroup = NoteHeadGroup.HEAD_DIAMOND : h.crossed && (1 === comboTabulatureDisplay.currentIndex ? durationIs(c, 1) || durationIs(c, 2) ? g.headGroup = NoteHeadGroup.HEAD_XCIRCLE : g.headGroup = NoteHeadGroup.HEAD_CROSS : 2 === comboTabulatureDisplay.currentIndex ? (console.log("Note duration: " + c.duration.numerator + " / " + c.duration.denominator), durationIs(c, 1) || durationIs(c, 2) ? 
          (addSharp2(g), setAccidentalVisible(!0)) : g.headGroup = NoteHeadGroup.HEAD_CROSS) : 3 === comboTabulatureDisplay.currentIndex ? (addSharp2(g), setAccidentalVisible(!0)) : g.headGroup = NoteHeadGroup.HEAD_CROSS);
        }
      }
    } else {
      if (a.Left) {
        for (c = a.Left.map(getMidiPitch), a = 0; a < e.length; a++) {
          b = e[a], c.includes(b.pitch) && (b.color = colorRed);
        }
      } else {
        console.warn("Invalid result for current chord " + b + ": " + JSON.stringify(a));
      }
    }
  }
}
function handleClickZugDruck(c) {
  var a = isReverseDirection();
  console.log("Starting translation: " + c + (a ? " reverse" : ""));
  var b = collectChords();
  if (0 === b.length) {
    console.warn("Keine Noten ausgew\u00e4hlt. Abbruch.");
  } else {
    if (b.length > maxChordLimit) {
      console.warn("Zu viele Noten ausgew\u00e4hlt. Abbruch.");
    } else {
      if (1 !== btnReverseDirection.state || isCurrentResultValid()) {
        if (a || !isCurrentResultValid()) {
          if (looksLikeGriffschrift(b) && !a) {
            console.warn("Markierte Noten sehen nach Griffschrift aus. Abbruch.");
            return;
          }
          alternativeIndex = 0;
          callApi(b, a, changeNotes(c, a));
          btnNextAlternative.enabled = !0;
        } else {
          alternativeIndex = c !== lastZD ? 0 : alternativeIndex + 1, changeNotes(c, !1)(b, lastResults);
        }
        invalidateResultsAfterTimeout();
        a && disableZDButtonsForTimeout();
      } else {
        console.log("Alternative Griffweisen durchzappen"), callApi(b, !0, toGriffschrift(c));
      }
    }
  }
}
function toGriffschrift(c) {
  return function(a, b) {
    a = extractZDResults(c, b);
    a = constructChordsFromNormalResults(a);
    console.log("Aus Normalnoten (die API zur\u00fcckgeliefert hat) werden jetzt MuseScore Notes erstellt, die dann in GS umgewandelt werden: " + JSON.stringify(a));
    alternativeIndex = 0;
    callApi(a, !1, changeNotes(c, !1));
  };
}


    Shortcut {
        sequence: "Alt+R"
        context: Qt.ApplicationShortcut
        onActivated: clickReverseDirection()
    }
    // Alt+D für Druck ist schon vergeben von Menü, deswegen J/K für Z/D als Alternative
    Shortcut {
        sequence: "Alt+J"
        context: Qt.ApplicationShortcut
        onActivated: {
            if (btnZug.enabled)
                clickZug()
        }
    }
    Shortcut {
        sequence: "Alt+K"
        context: Qt.ApplicationShortcut
        onActivated: {
            if (btnDruck.enabled)
                clickDruck()
        }
    }

    MessageDialog {
        id: errorDialog
        title: "Fehler"
        icon: StandardIcon.Critical
        function show(errorMessage) {
            text = errorMessage
            open()
        }
    }

    Timer {
      id: invalidateResultsTimer
      interval: 100000
      running: false
      repeat: false
      //singleShot: true
      onTriggered: invalidateCurrentResults()
    }

    Timer {
      id: disableZDButtonsTimer
      interval: 2000
      running: false
      repeat: false
      //singleShot: true
      onTriggered: enableZDButtons(true)
    }

    function enableZDButtons(yes) {
        console.log("zd buttons enabled: " + yes)
        btnZug.enabled = yes
        btnDruck.enabled = yes
    }

    function invalidateCurrentResults() {
        btnNextAlternative.enabled = false
        lastResults = null
    }

    function isCurrentResultValid() {
        return !!lastResults
    }

    function disableZDButtonsForTimeout() {
        enableZDButtons(false)
        disableZDButtonsTimer.restart()
    }

    function invalidateResultsAfterTimeout() {
        invalidateResultsTimer.restart()
    }

    function clickZug() {
        handleClickZugDruck('zug')
    }
    function clickDruck() {
        handleClickZugDruck('druck')
    }
    function isReverseDirection() {
        //return comboDirection.currentIndex !== 0
        return btnReverseDirection.state === 2
    }
    function clickNextAlternative() {
        // Only for Normalnotation → GS
        if (!isCurrentResultValid() || isReverseDirection()) {
            error.warn("btnNextAlternative should be disabled. Aborting.")
            return
        }
        var chords = collectChords()
        alternativeIndex += 1
        changeNotes(lastZD, false)(chords, lastResults)
    }

    function clickReverseDirection() {
        var state = btnReverseDirection.state
        state = (state + 1) % 3
        btnReverseDirection.state = state
        btnReverseDirection.text = btnReverseDirection.texts[state]
        btnReverseDirection.background.color = (state === 2) ? colorDirectionReverse : colorButtonNormal
    }

    GridLayout {
        //anchors.fill: parent
        Layout.fillWidth: true
        columns: 1
        rowSpacing: 1

        Text {
            text: "Instrument (anzeigen)"
            horizontalAlignment: Text.AlignHCenter
            Layout.fillWidth: true
            MouseArea {
                anchors.fill: parent
                onClicked: {
                    // TODO Directly open first selected chord? note names must be in German for this!
                    var model = comboModel.currentKey()
                    var chords = collectChords()
                    //let chord = chords.length > 0 ? chords[0] : []
                    //console.log(chords)
                    var noteNames = []
                    //for (let i = 0; i < chord.notes.length; i++) {
                    //    var note = chord.notes[i]
                    //    var name = germanNoteNames.find(([n, _]) => n === note)
                    //    if (name) {
                    //        noteNames.push(name[1])
                    //    }
                    //}
                    console.log(noteNames.join(' '))
                    Qt.openUrlExternally(apiUrl + '?' +
                                         queryStringArg('model', model, true) +
                                         (noteNames.length ? queryStringArg('notes', noteNames.join(' '), false) : '') +
                                         (txtLicenseKey.text ? queryStringArg('license', txtLicenseKey.text) : ''))
                }
            }
        }
        ComboBox {
            id: comboModel
            Layout.fillWidth: true
            textRole: "value"
            model: ListModel {
                // Generated using Nn2GS.hs:
                // putStrLn $ unlines $ map (\x -> "ListElement { key: \"" ++ map toLower x ++ "\"; value: \"" ++ x ++ "\" }") $ map modelId listOfInstruments
                ListElement { key: "adgc50"; value: "ADGC50" }
                ListElement { key: "besasdes50"; value: "BEsAsDes50" }
                ListElement { key: "gcfb50"; value: "GCFB50" }
                ListElement { key: "club_cf"; value: "Club_CF" }
                ListElement { key: "knopfakkordeon_b"; value: "Knopfakkordeon_B" }
            }
            function currentKey() {
                return model.get(currentIndex).key
            }
        }
        Text {
            text: "Aktuelle Dur-Tonart"
            horizontalAlignment: Text.AlignHCenter
            Layout.fillWidth: true
        }

        SpinBox {
            width: parent.width
            Layout.fillWidth: true
            id: spinnerTonart
            value: 8
            from: 0
            to: 20
            editable: true
            //validator: QRegExpValidator(/.*/) // tonarten.join('|') + ^$ anchors?
            validator: null
            textFromValue: function(value, locale) {
                return tonarten[value][1]
            }
            valueFromText: function(text, locale) {
                var i = tonarten.indexOf(text)
                if (i >= 0) return i
                return 0
            }
        }

        ComboBox {
            id: comboDirection
            Layout.fillWidth: true
            visible: false // in favor of btnReverseDirection
            textRole: "value"
            model: ListModel {
                  ListElement { value: "Normal → GS" }
                  ListElement { value: "GS-Alternativen" }
                  ListElement { value: "GS → Normal" }
            }
            onCurrentIndexChanged: {
                invalidateCurrentResults()
            }
        }
        Button {
            id: btnReverseDirection
            Layout.fillWidth: true
            readonly property var texts: ["Normal → GS", "GS-Alternativen", "GS → Normal"]
            //background.color: colorButtonNormal
            text: "Reverse"
            property int state: 0
            checkable: false // checked state doesn't work reliably with double clicks
            checked: false
            // Handle both events because on double click, click AND
            // double click are fired which causes a faster switching:
            onClicked: clickReverseDirection()
            //onDoubleClicked: clickReverseDirection() <- does not work on Windows
        }

        Button {
            text: "Zug"
            id: btnZug
            Layout.fillWidth: true
            //background.color: colorButtonNormal
            onClicked: clickZug()
        }
        Button {
            text: "Druck"
            id: btnDruck
            Layout.fillWidth: true
            //background.color: colorButtonNormal
            onClicked: clickDruck()
        }
        Button {
            text: "Nächste Alternative"
            Layout.fillWidth: true
            //background.color: colorButtonNormal
            visible: false
            enabled: false
            id: btnNextAlternative
            onClicked: clickNextAlternative()
        }

        Text {
            text: "Einstellungen"
            horizontalAlignment: Text.AlignHCenter
            Layout.fillWidth: true
        }
        ComboBox {
            id: comboTabulatureDisplay
            Layout.fillWidth: true
            textRole: "value"
            enabled: true
            currentIndex: 0
            model: ListModel {
                  ListElement { value: "Notenköpfe in Kreuzform" }
                  ListElement { value: "Kreuz in Halben/Ganzen" }   // Johannes Servi
                  //ListElement { value: "Kreuz vor Halben/Ganzen" }  // Matthias Pürner
                  //ListElement { value: "Kreuz vor allen" }          // Rosenzopf
            }
        }
        TextField {
            text: ""
            id: txtLicenseKey
            Layout.fillWidth: true
            visible: false
            placeholderText: "Lizenz-Schlüssel eingeben…"
            Keys.onPressed: {
                if (text.trim() === '' && event.key === Qt.Key_Return) {
                    console.log("Opening URL in browser.")
                    Qt.openUrlExternally(apiUrl + '#faq-kosten')
                    text = ''
                }
            }
            // Use local storage to save the key for the next session
            // https://doc.qt.io/qt-5/qtquick-localstorage-qmlmodule.html
        }
     }
  }
